/*
  Sedov Taylor explosion
*/

#include <cassert>
#include <iostream>
#include <fstream>
#include <vector>
#include "srhd_sim.hpp"
#include "spatial_distribution.hpp"
#include "ideal_gas.hpp"
//#include "imgrs.hpp"
#include "hll.hpp"
#include "pcm.hpp"
#include "rigid_wall.hpp"
#include "trigger.hpp"
#include "rubric.hpp"
#include "consecutive_snapshots.hpp"
#include "main_loop.hpp"
#include <fenv.h>

using namespace std;

namespace {
  double get_position_max_pressure(SRHDSimulation const& sim)
  {
    double pmax = sim.getHydroSnapshot().cells[0].Pressure;
    double rpmax = sim.GetCellCentre(0);
    for(size_t i=0;i<sim.getHydroSnapshot().cells.size();i++){
      Primitive p = sim.getHydroSnapshot().cells[i];
      if(pmax<p.Pressure){
	pmax = p.Pressure;
	rpmax = sim.GetCellCentre(i);
      }
    }
    return rpmax;
  }

  void WriteVectors2File(vector<double> const& v1,
			 vector<double> const& v2,
			 string const& fname)
  {
    ofstream f;
    f.open(fname.c_str());
    for(size_t i=0;i<v1.size();i++){
      f << v1[i] << " " << v2[i] << endl;
    }
    f.close();
  }

  vector<double> my_logspace(double low_exp,
			     double high_exp,
			     size_t num)
  {
    const vector<double> dummy = linspace(low_exp, high_exp, num);
    vector<double> res(num,0);
    for(size_t i=0;i<num;++i)
      res.at(i) = pow(10,dummy.at(i));
    return res;
  }

  class SimData
  {
  public:

    SimData(void):
      eos_(4./3.),
      rs_(eos_),
      bc_(rs_),
      sr_(),
      geometry_(),
      sim_(
	   my_logspace(-5,0,1000),
	   //	   logspace(0,1,1000,1.01),
	   //	   linspace(0,1,2000),
	   Uniform(1),
	   Step(1e5, 1e-9, 1e-4),
	   Uniform(0),
	   bc_, bc_,
	   eos_,
	   rs_,
	   sr_,
	   geometry_) {}

    SRHDSimulation& getSim(void)
    {
      return sim_;
    }

  private:
    IdealGas eos_;
    HLL rs_;
    RigidWall bc_;
    PCM sr_;
    const Planar geometry_;
    SRHDSimulation sim_;
  };

  class ShockFrontTracker: public DiagnosticFunction
  {
  public:
    
    ShockFrontTracker(const string& fname):
      fname_(fname), positions_(), times_() {}

    void operator()(const SRHDSimulation& sim) const
    {
      times_.push_back(sim.GetTime());
      positions_.push_back(get_position_max_pressure(sim));
    }

    ~ShockFrontTracker(void)
    {
      WriteVectors2File(times_, positions_, fname_);
    }
    
  private:
    const string fname_;
    mutable vector<double> positions_;
    mutable vector<double> times_;
  };

  class RegularIntervals: public Trigger
  {
  public:
    RegularIntervals(double dt):
      dt_(dt),
      t_next_(0),
      count_(0) {}

    bool operator()(const SRHDSimulation& sim)
    {
      const double time = sim.GetTime();
      if(time<t_next_)
	return false;
      t_next_ += dt_;
      count_++;
      return true;
    }

    int getCount(void) const
    {
      return count_;
    }

  private:
    const double dt_;
    mutable double t_next_;
    mutable int count_;
  };

  void my_main_loop(SRHDSimulation& sim)
  {
    const double tf = 1.0;
    //    ShockFrontTracker sft("rpmax.txt");
    SafeTimeTermination stt(tf, 1e9);
    RegularIntervals ri(tf/100);
    Rubric rubric("output/snapshot_",".h5");
    MultipleDiagnostics diag;
    diag.diag_list.push_back(new ShockFrontTracker("rpmax.txt"));
    diag.diag_list.push_back(new ConsecutiveSnapshots(ri, rubric));
    main_loop(sim, stt, &SRHDSimulation::TimeAdvance, diag);
  }
}

int main()
{
  feenableexcept(FE_INVALID   | 
		 FE_DIVBYZERO | 
		 FE_OVERFLOW  | 
		 FE_UNDERFLOW);

  my_main_loop(SimData().getSim());

  ofstream("test_terminated_normally.res").close();
  return 0;
}
